
% Compile code if necessary
try
    dartel3('smalldef',single(zeros(1,1,1,3)));
catch
    mex dartel3.c optimizer3d.c diffeo3d.c -O
end
addpath(fullfile(fileparts(which('SHOOT_DEMO'))),'spm');

demo_setup

d        = size(F{1});
d        = d(1:2);
prm      = [0  1 1 [0.025 0.1 1e-4]];
int_args = [16 2 2];
bs_args  = [3 3 0 1 1 1];
[id{1:3}]= ndgrid(1:d(1),1:d(2),1);
phi      = cat(3,id{1:2});

dat = struct('f',F, 'v0',zeros([d,2]), 'v1',[], 'phi',phi, 'dj',ones(d));
for i=1:numel(dat)
    dat(i).f = spm_bsplinc(dat(i).f,bs_args);
end
Eall = Inf;
scal = 1;
smo  = 8;
dec  = 0.7;
for iter=1:30,

    mu = zeros(d);
    n = zeros(d);
    for i=1:numel(dat)
        [phi,J]    = shoot2d(dat(i).v0,prm,int_args); % Geodesic shooting
        dat(i).dj  = J(:,:,1,1).* J(:,:,2,2) -  J(:,:,1,2).* J(:,:,2,1); % Determinants
        dat(i).phi = phi;
        f1         = spm_bsplins(dat(i).f,phi(:,:,1),phi(:,:,2),id{3},bs_args);
        mu         = mu + f1.*dat(i).dj; % Sufficient stats for mean
        n          = n  + dat(i).dj;

        % Show some pictures
        subplot(3,numel(dat),i  );
        imagesc(f1'); axis image xy off;
        subplot(3,numel(dat),i+numel(dat));
        imagesc(dat(i).dj'); axis image xy off;
        subplot(3,numel(dat),i+numel(dat)*2);
        sk=4;
        plot(phi(:,1:sk:end,1),phi(:,1:sk:end,2),'k-',phi(1:sk:end,:,1)',phi(1:sk:end,:,2)','b-')
        axis image xy off
        drawnow
    end

    % Re-compute template
    mu          = spm_conv(mu,smo)./spm_conv(n,smo);
    smo         = smo*dec; % Decrease smoothing
    mu          = spm_bsplinc(mu,bs_args);
    [mu1,d1,d2] = spm_bsplins(mu, id{:},bs_args);

    avg_v0 = zeros([d,2]);
    Eprev  = Eall;
    Eall   = 0;
    for i=1:numel(dat),

        f1        = spm_bsplins(dat(i).f,dat(i).phi(:,:,1),dat(i).phi(:,:,2),id{3},bs_args);
        tmp       = dat(i).dj.*(f1-mu1);
        Elik      = 0.5*sum(tmp(:).^2); % Likelihood term

        g         = zeros([d,2]); % First derivatives
        g(:,:,1)  = tmp.*d1;
        g(:,:,2)  = tmp.*d2;

        H         = zeros([d,3]); % Second derivatives
        H(:,:,1)  = dat(i).dj.*d1.*d1;
        H(:,:,2)  = dat(i).dj.*d2.*d2;
        H(:,:,3)  = dat(i).dj.*d1.*d2;

        u0        = vel2mom2d(dat(i).v0,prm);     % Initial momentum
        Eprior    = 0.5*sum(u0(:).*dat(i).v0(:)); % Prior term
        closeness = sum((g(:)+u0(:)).^2)/(sum(g(:).^2)+eps); % Measure of convergence
        fprintf('%2d\t%g\t%g\t%g\t%g\n', iter, ...
            Elik/prod(d), Eprior/prod(d), (Eprior+Elik)/prod(d), ...
            closeness);
        Eall      = Eall + Elik+Eprior; % Objective function for all

        dat(i).v0 = dat(i).v0 - scal*multigrid2d(H, g+u0, [prm(1:5) prm(6) 4 4]); % Gauss-Newton update

        avg_v0    = avg_v0 + dat(i).v0; % For mean correcting initial velocities
    end

    fprintf('   *** %g \t %g ***\n', Eall, scal);
    % Updates too big/small?
    if Eall>Eprev, scal = scal*0.5; else scal = scal*1.1; end
    scal = max(min(scal,1.0),0.1);

    % Total initial velocity should be zero (Khan & Beg)
    avg_v0 = avg_v0/numel(dat);
    for i=1:numel(dat),
        dat(i).v0 = dat(i).v0 - avg_v0;
    end

    % Finished?
    if iter==1, Eit1 = Eall; end
    if abs(Eall-Eprev)<Eit1*1e-5, break; end

end

% Generate a few more pictures
fpics = figure; colormap(1-gray)
fwarp = figure; colormap(1-gray)
fdet  = figure; colormap(1-gray)

for i1=1:numel(dat)
    [phi,J,v1] = shoot2d(dat(i1).v0,prm,int_args);
    dj      =  J(:,:,1,1).* J(:,:,2,2) -  J(:,:,1,2).* J(:,:,2,1);
    for i2=1:numel(dat)
        ppos    = {numel(dat),numel(dat),(i2-1)*numel(dat)+i1};
        [bphi,bJ,bv1,biphi,biJ] = shoot2d(dat(i2).v0,prm,int_args);
        [y1,J1] = comp2d(phi,biphi,J,biJ);
        dj1     = J1(:,:,1,1).* J1(:,:,2,2) -  J1(:,:,1,2).* J1(:,:,2,1);

        figure(fwarp); subplot(ppos{:});
        sk      = 4;
        plot(y1(:,1:sk:end,1),y1(:,1:sk:end,2),'k-',y1(1:sk:end,:,1)',y1(1:sk:end,:,2)','b-')
        axis image xy off

        figure(fpics); subplot(ppos{:});
        [f1,df1,df2] = spm_bsplins(double(dat(i1).f),y1(:,:,1),y1(:,:,2),id{3},bs_args);
        imagesc(f1'); axis image xy off

        figure(fdet); subplot(ppos{:});
        imagesc(dj1'); axis image xy off; set(gca,'Clim',[0 6]);
    end
end

